from __future__ import annotations

import builtins
import inspect
from typing import Any, cast, Optional, Tuple

from typing_extensions import Annotated, get_args, get_origin

from ._injection import ArgDependency
from .annotations import (AntidoteAnnotation, From, FromArg, Get, INJECT_SENTINEL)
from .injection import Arg
from .marker import InjectMeMarker, Marker
from .._internal import API
from .._internal.argspec import Argument


@API.private
def extract_annotated_dependency(type_hint: object) -> object:
    origin = get_origin(type_hint)

    # Dependency explicitly given through Annotated (PEP-593)
    if origin is Annotated:
        args = get_args(type_hint)
        antidote_annotations = [a
                                for a in getattr(type_hint, "__metadata__", tuple())
                                if isinstance(a, AntidoteAnnotation)]
        if len(antidote_annotations) > 1:
            raise TypeError(f"Multiple AntidoteAnnotation are not supported. "
                            f"Found {antidote_annotations}")
        elif antidote_annotations:
            annotation: AntidoteAnnotation = antidote_annotations[0]
            if annotation is INJECT_SENTINEL:
                return args[0]
            elif isinstance(annotation, Get):
                return annotation.dependency
            elif isinstance(annotation, From):
                return args[0] @ annotation.source
            else:
                raise TypeError(f"Annotation {annotation} cannot be used"
                                f"outside of a function.")
        else:
            return args[0]

    return type_hint


_BUILTINS_TYPES = {e for e in builtins.__dict__.values() if isinstance(e, type)}


@API.private
def extract_annotated_arg_dependency(argument: Argument) -> object:
    type_hint, origin, args = _extract_type_hint(argument)

    # Dependency explicitly given through Annotated (PEP-593)
    if origin is Annotated:
        antidote_annotations = [a
                                for a in type_hint.__metadata__
                                if isinstance(a, AntidoteAnnotation)]
        if len(antidote_annotations) > 1:
            raise TypeError(f"Multiple AntidoteAnnotation are not supported. "
                            f"Found {antidote_annotations}")
        elif antidote_annotations:
            if isinstance(argument.default, Marker):
                raise TypeError("Cannot use a Marker with an Antidote annotation.")

            # If antidote annotation, no additional check is done we just return
            # what was specified.
            annotation: AntidoteAnnotation = antidote_annotations[0]
            if annotation is INJECT_SENTINEL:
                return args[0]
            elif isinstance(annotation, Get):
                return annotation.dependency
            elif isinstance(annotation, From):
                return args[0] @ annotation.source
            elif isinstance(annotation, FromArg):
                arg = Arg(argument.name,
                          argument.type_hint,
                          argument.type_hint_with_extras)
                return annotation.function(arg)  # type: ignore
            else:
                raise TypeError(f"Unsupported AntidoteAnnotation, {type(annotation)}")

    if isinstance(argument.default, Marker):
        from .._constants import LazyConst

        marker = argument.default
        dependency: object
        if isinstance(marker, Get):
            dependency = marker.dependency
        elif isinstance(marker, LazyConst):
            dependency = cast(object, marker)
        elif isinstance(marker, InjectMeMarker):
            if not is_valid_class_type_hint(type_hint):
                raise TypeError(
                    f"Cannot use marker @inject.me with non class type hint: {type_hint}")
            if marker.source is not None:
                dependency = Get(type_hint, source=marker.source).dependency
            else:
                dependency = type_hint
        else:
            raise TypeError("Custom Marker are NOT supported.")

        return ArgDependency(dependency, optional=argument.is_optional)

    return None


@API.private
def extract_auto_provided_arg_dependency(argument: Argument) -> Optional[type]:
    type_hint, origin, args = _extract_type_hint(argument)
    dependency = type_hint

    if origin is Annotated:
        antidote_annotations = [a
                                for a in type_hint.__metadata__
                                if isinstance(a, AntidoteAnnotation)]
        if not antidote_annotations:
            dependency = args[0]

    if is_valid_class_type_hint(dependency):
        return cast(type, dependency)

    return None


@API.private
def is_valid_class_type_hint(type_hint: object) -> bool:
    return (getattr(type_hint, '__module__', '') != 'typing'
            and type_hint not in _BUILTINS_TYPES
            and (isinstance(type_hint, type) and inspect.isclass(type_hint)))


@API.private
def _extract_type_hint(argument: Argument) -> Tuple[Any, object, Tuple[object, ...]]:
    type_hint = argument.type_hint_with_extras
    origin = get_origin(type_hint)
    args = get_args(type_hint)

    # Optional
    if argument.is_optional:
        type_hint = args[0] if isinstance(None, args[1]) else args[0]
        origin = get_origin(type_hint)
        args = get_args(type_hint)

    return type_hint, origin, args



from typing import Hashable

from typing_extensions import final

from .._internal import API
from ..core import Container, DependencyDebug, DependencyValue, StatelessProvider
from ..core.exceptions import DebugNotAvailableError


@API.private
class Lazy:
    def __antidote_debug_info__(self) -> DependencyDebug:
        raise DebugNotAvailableError()

    def __antidote_provide__(self, container: Container) -> DependencyValue:
        raise NotImplementedError()  # pragma: no cover


@API.private
@final
class LazyProvider(StatelessProvider[Lazy]):
    def exists(self, dependency: Hashable) -> bool:
        return isinstance(dependency, Lazy)

    def debug(self, dependency: Lazy) -> DependencyDebug:
        return dependency.__antidote_debug_info__()

    def provide(self, dependency: Lazy, container: Container
                ) -> DependencyValue:
        return dependency.__antidote_provide__(container)



import pytest

from antidote._internal.utils.immutable import FinalImmutable, Immutable


@pytest.fixture(params=[Immutable, FinalImmutable])
def cls(request):
    return request.param


def test_immutable_meta(cls: type):
    class A(cls):
        __slots__ = ('value',)

    class B(cls):
        __slots__ = ('__value',)

        def get(self):
            return self.__value

    x = object()
    a = A(x)
    b = B(x)

    assert a.value is x
    assert b.get() is x


def test_invalid_immutable(cls: type):
    with pytest.raises(TypeError, match=".*slots.*"):
        class A(cls):
            pass


def test_immutability(cls: type):
    class A(cls):
        __slots__ = ('x', 'y')

        def __init__(self, x, y):
            super().__init__(x=x, y=y)

    x = object()
    y = object()

    a = A(x, y)
    assert a.x is x
    assert a.y is y

    with pytest.raises(AttributeError):
        a.x = 1

    with pytest.raises(AttributeError):
        a.unknown = 1



from typing import TypeVar

from typing_extensions import Annotated

T = TypeVar('T')


def public(x: T) -> T:
    """
    Objects marked with this decorator are considered to be in the public API.
    Breaking changes will be avoided and will be taken into account in the
    semantic versioning.
    """
    return x  # pragma: no cover


def deprecated(x: T) -> T:
    """
    Objects marked with this decorator are considered to be deprecated. Use the documented
    alternative if it exists, as this object will be removed in a future release.
    """
    return x  # pragma: no cover


def experimental(x: T) -> T:
    """
    Similar to public, they're part of the public API *BUT* they're not stable.
    As such there's a good chance they may be changed or removed in the next release. If
    you're relying on an experimental feature consider opening an issue with your use case
    to advocate its migration to public !
    """
    return x  # pragma: no cover


def private(x: T) -> T:
    """
    Only for internal use. They are NOT part of the public API, and as such may
    change without warning in later versions. If you rely on private APIs, please open
    an issue.
    """
    return x  # pragma: no cover


Deprecated = Annotated[T, "deprecated argument"]



import weakref
from typing import Callable, Optional

from typing_extensions import final

from ._internal import API
from ._internal.utils import debug_repr, FinalImmutable, short_id
from ._lazy import (LazyCallWithArgsKwargs, LazyMethodCallDependency,
                    LazyMethodCallWithArgsKwargs)
from ._providers import Lazy
from .core import Container, DependencyDebug, DependencyValue, Scope
from .service import Service
from .utils import validated_scope


@API.public
@final
class LazyCall(FinalImmutable, Lazy):
    """
    Declares the result of a function call as a depdency.

    .. doctest:: lazy_func

        >>> from antidote import LazyCall, world
        >>> def f(x, y):
        ...     return x + y
        >>> Computation = LazyCall(f)(2, y=3)
        >>> world.get(Computation)
        5
        >>> user = 'John'
        >>> def hello():
        ...     return f"Hello {user}"
        >>> HelloUser = LazyCall(hello, singleton=False)
        >>> world.get(HelloUser)
        'Hello John'
        >>> user = "Adam"
        >>> world.get(HelloUser)
        'Hello Adam'

    """
    __slots__ = ('func', '_scope')
    func: Callable[..., object]
    _scope: Optional[Scope]

    def __init__(self,
                 func: Callable[..., object],
                 *,
                 singleton: Optional[bool] = None,
                 scope: Optional[Scope] = Scope.sentinel()) -> None:
        """
        Args:
            func: Function to lazily call, any (keyword-)arguments given by the returned
                :py:class:`~.LazyCall` itself will be propagated.
            singleton: Whether the lazy dependency is a singleton or not. If yes,
                the function will be called at most once and the result re-used. Mutually
                exclusive with :code:`scope`. Defaults to :py:obj:`True`.
            scope: Scope of the dependency. Mutually exclusive with :code:`singleton`.
                The scope defines if and how long the returned dependency will be
                cached. See :py:class:`~.core.container.Scope`. Defaults to
                :py:meth:`~.core.container.Scope.singleton`.
        """
        if not callable(func):
            raise TypeError(f"func must be a callable, not {type(func)}")
        super().__init__(func, validated_scope(scope,
                                               singleton,
                                               default=Scope.singleton()))

    def __call__(self, *args: object, **kwargs: object) -> LazyCallWithArgsKwargs[object]:
        """
        All argument are passed on to the lazily called function.
        """
        return LazyCallWithArgsKwargs(self.func, self._scope, args, kwargs)

    @API.private
    def __antidote_debug_repr__(self) -> str:
        s = f"Lazy: {debug_repr(self.func)}()"
        if self._scope is not None:
            s += f"  #{short_id(self)}"
        return s

    @API.private
    def __antidote_debug_info__(self) -> DependencyDebug:
        return DependencyDebug(self.__antidote_debug_repr__(),
                               scope=self._scope,
                               wired=[self.func])

    @API.private
    def __antidote_provide__(self, container: Container) -> DependencyValue:
        return DependencyValue(self.func(), scope=self._scope)


@API.deprecated
@final
class LazyMethodCall(FinalImmutable):
    """
    Similar to :py:class:`~.LazyCall` but adapted to methods within a class definition.
    It can only be used with a :py:class:`.Service` subclass.

    :py:class:`~.LazyMethodCall` behaves in a similar way as constants in
    :py:class:`.Constants`. When accessed as class attributes, the dependency is returned.
    But on an instance, the actual method call result is returned.

    .. doctest:: lazy_method

        >>> from antidote import LazyMethodCall, Service, world
        >>> class Api(Service):
        ...     def __init__(self, host: str = 'localhost'):
        ...         self.host = host
        ...
        ...     def query(self, url: str = '/status'):
        ...         return f"requested {self.host}{url}"
        ...
        ...     status = LazyMethodCall(query, singleton=False)
        ...     conf = LazyMethodCall(query)('/conf')
        >>> world.get(Api.conf)
        'requested localhost/conf'
        >>> world.get(Api.status)
        'requested localhost/status'
        >>> # For ease of use, accessing the dependency through a instance will simply
        ... # call the method without passing through Antidote. Useful for tests typically
        ... Api(host='example.com').status
        'requested example.com/status'

    .. note::

        Consider using :py:class:`~.extension.constants.Constants` if you only declare
        constants.

    """
    __slots__ = ('_method_name', '_scope', '__cache_attr')
    _method_name: str
    _scope: Optional[Scope]
    __cache_attr: str

    def __init__(self,
                 method: Callable[..., object],
                 *,
                 singleton: Optional[bool] = None,
                 scope: Scope = Scope.sentinel()) -> None:
        """
        Args:
            method: Method name or the method itself that must be called.
            singleton: Whether the lazy dependency is a singleton or not. If yes,
                the function will be called at most once and the result re-used. Mutually
                exclusive with :code:`scope`. Defaults to :py:obj:`True`.
            scope: Scope of the dependency. Mutually exclusive with :code:`singleton`.
                The scope defines if and how long the returned dependency will be
                cached. See :py:class:`~.core.container.Scope`. Defaults to
                :py:meth:`~.core.container.Scope.singleton`.
        """
        if not callable(method):
            raise TypeError("method must be a method or its name")
        super().__init__(
            method.__name__,
            validated_scope(scope, singleton, default=Scope.singleton()),
            f"__antidote_dependency_{hex(id(self))}"
        )

    def __call__(self, *args: object, **kwargs: object) -> LazyMethodCallWithArgsKwargs:
        """
        All argument are passed on to the lazily called method.
        """
        return LazyMethodCallWithArgsKwargs(self._method_name,
                                            self._scope,
                                            args,
                                            kwargs)

    def __str__(self) -> str:
        s = f"Lazy Method: {self._method_name}()"
        if self._scope is not None:
            s += f"  #{short_id(self)}"
        return s

    def __get__(self, instance: object, owner: type) -> object:
        if not issubclass(owner, Service):
            raise RuntimeError("LazyMethod can only be used on a Service subclass.")

        if instance is None:
            try:
                return getattr(owner, self.__cache_attr)
            except AttributeError:
                dependency = LazyMethodCallDependency(self, weakref.ref(owner))
                setattr(owner, self.__cache_attr, dependency)
                return dependency
        return getattr(instance, self._method_name)()



import pytest

from antidote._internal.stack import DependencyStack
from antidote.exceptions import DependencyCycleError


class Service:
    pass


class CustomException(Exception):
    pass


def test_instantiating():
    stack = DependencyStack()
    a = object()
    b = object()

    with stack.instantiating(a):
        with stack.instantiating(b):
            pass

        # stack not have b anymore
        with stack.instantiating(b):
            pass

        with pytest.raises(DependencyCycleError):
            with stack.instantiating(a):
                pass

    # stack should be clean after cycle error
    with stack.instantiating(a):
        pass

    with pytest.raises(CustomException):
        with stack.instantiating(a):
            raise CustomException()

    # stack should be clean after user error
    with stack.instantiating(a):
        pass


def test_is_empty():
    stack = DependencyStack()
    a = object()
    b = object()

    assert stack.depth == 0

    with stack.instantiating(a):
        assert stack.depth == 1

        with stack.instantiating(b):
            assert stack.depth == 2

        assert stack.depth == 1

    assert stack.depth == 0


def test_to_list():
    stack = DependencyStack()
    a = object()
    b = object()

    assert stack.to_list() == []

    with stack.instantiating(a):
        assert stack.to_list() == [a]

        with stack.instantiating(b):
            assert stack.to_list() == [a, b]
            stack_copy = stack.to_list()
            stack_copy.append('dummy')
            # should not have impacted the actual stack
            assert stack.to_list() == [a, b]

        assert stack.to_list() == [a]

    assert stack.to_list() == []
